"""Runs stats describe on data. Prints data in a markdown table."""

__copyright__ = "Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved."
__author__ = "DV Klopfenstein"

import sys
import math
import numpy as np
import scipy.stats as stats

class StatsDescribe(object):
    """Describe summary statistics for a list of numbers in a markdown table."""

    fmt = "{name:10} | {qty:13} | {range:20} | {25th percentile:>15} | " \
          "{median:>8} | {75th percentile:>15} | {mean:>8} | {stddev:>}\n"

    def __init__(self, desc, fmtstr="{:>8.2e}"):
        self.desc = desc
        self.fmtstr = fmtstr

    def prt_hdr(self, prt=sys.stdout, name="name       "):
        """Print stats header in markdown style."""
        hdr = "{NAME} | # {ITEMS:11} | range                | 25th percentile | " \
              "  median | 75th percentile |     mean | stddev\n".format(NAME=name, ITEMS=self.desc)
        div = "{DASHES}|---------------|----------------------|" \
              "-----------------|----------|-----------------|----------|-------\n".format(
                  DASHES='-'*(len(name)))
        prt.write(hdr)
        prt.write(div)

    def prt_data(self, name, vals, prt=sys.stdout):
        """Print stats data in markdown style."""
        fld2val = self.get_fld2val(name, vals)
        prt.write(self.fmt.format(**fld2val))
        return fld2val

    def getstr_data(self, name, vals):
        """Return stats data string in markdown style."""
        fld2val = self.get_fld2val(name, vals)
        return self.fmt.format(**fld2val)

    def get_fld2val(self, name, vals):
        """Describe summary statistics for a list of numbers."""
        if vals:
            return self._init_fld2val_stats(name, vals)
        return self._init_fld2val_null(name)

    def _init_fld2val_stats(self, name, vals):
        """Return statistics on values."""
        vals_stats = stats.describe(vals)
        stddev = math.sqrt(vals_stats[3]) # stats variance
        p25 = np.percentile(vals, 25)
        p50 = np.percentile(vals, 50) # median
        p75 = np.percentile(vals, 75)
        fld2val = {
            'name':name,
            'qty'.format(ITEMS=self.desc):vals_stats[0], # stats nobs
            'range':self._get_str_range(vals_stats),
            '25th percentile':p25,
            'median':p50,
            '75th percentile':p75,
            'mean':vals_stats[2], # stats mean
            'stddev':stddev}
        fmtflds = set(['25th percentile', 'median', '75th percentile', 'mean', 'stddev'])
        mkint = "," in self.fmtstr
        for key, val in fld2val.items():
            if key in fmtflds:
                if mkint:
                    val = int(round(val))
                fld2val[key] = self.fmtstr.format(val)
        return fld2val

    def _init_fld2val_null(self, name):
        """Return empty fields if there are no values."""
        return {
            'name':name,
            'qty'.format(ITEMS=self.desc):0,
            'range':"",
            '25th percentile':"",
            'median':"",
            '75th percentile':"",
            'mean':"",
            'stddev':""}

    def _get_str_range(self, vals_stats):
        """Return a string containing the range of values."""
        minmax = vals_stats[1] # stats minmax
        minval = self.fmtstr.format(minmax[0])
        maxval = self.fmtstr.format(minmax[1])
        return '{A} to {B:6>}'.format(A=minval, B=maxval)

# Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved.



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

""" Maps gene associations to a 'slim' ontology.

    This roughly implements the functionality of the perl script map2slim.
    See: [http://search.cpan.org/~cmungall/go-perl/scripts/map2slim]

    For a description of GO Slims look here:
    http://geneontology.org/GO.slims.shtml

    For now this does not implement Bucket Terms.
"""

from .obo_parser import GODag


def mapslim(go_term, go_dag, goslim_dag):
    """ Maps a GO term (accession) to it's GO slim terms.

        Parameters:
        - go_term: the accession to be mapped to the slim terms
        - go_dag: the (full) Gene Ontology DAG
        - goslim_dag: the GO Slim DAG

        Returns:
            Two sets:
            direct_ancestors, all_ancestors
        - direct_ancestors: The direct ancestors of the given term that are in
                            the GO Slim. Those are the terms that are not
                            covered by earlier ancestors of the GO Slims in
                            _any_ path (from bottom to top).
        - all_ancestors:    All ancestors of the given term that are part of
                            the GO-Slim terms.

    """
    # check parameters
    if not isinstance(go_dag, GODag):
        raise TypeError("go_dag must be an instance of GODag")
    if not isinstance(goslim_dag, GODag):
        raise TypeError("goslim_dag must be an instance of GODag")
    if go_term not in go_dag:
        raise ValueError("go_term must be an accession that is in the go_dag")

    all_ancestors = set()
    covered_ancestors = set()

    # get all paths for the term in the go_dag
    paths = go_dag.paths_to_top(go_term)
    for path in paths:
        # the next loop needs to run bottom->up, i.e. from the go_term item to
        # the root, thus we need to reverse the list prior to iteration
        path.reverse()

        got_leaf = False
        for term in path:
            if term.id in goslim_dag:
                all_ancestors.add(term.id)
                if got_leaf:
                    covered_ancestors.add(term.id)
                got_leaf = True

    # get the direct ancestors, i.e. those that are not covered by a earlier
    # ancestor of the GO-Slim in _any_ path (in bottom->top order)
    direct_ancestors = all_ancestors - covered_ancestors
    return direct_ancestors, all_ancestors



"""Used to operate on a sub-graph of a larger GO DAG."""

from collections import defaultdict

__copyright__ = "Copyright (C) 2016-2018, DV Klopfenstein, H Tang, All rights reserved."
__author__ = "DV Klopfenstein"

class GODagSmall(object):
    """Contains a sub-graph of the original obo from geneontology.org."""

    def __init__(self):
        # Sub-graph of input DAG. User has option of using a subset of children/parents
        self.go_sources = None
        self.go2obj = {}
        self.p_from_cs = defaultdict(set) # GO ids for child->parents
        self.c_from_ps = defaultdict(set) # GO ids for parent->children

    def num_goterms(self):
        """Return the number of GO terms in this GO DAG subgraph."""
        return len(self.go2obj)

    def get_edges(self):
        """Get the directed edges from GO term to GO term."""
        edge_from_to = []
        for parent, children in self.p_from_cs.items():
            for child in children:
                edge_from_to.append((child, parent))
        for parent, children in self.c_from_ps.items():
            for child in children:
                edge_from_to.append((child, parent))
        return edge_from_to
        
   
# Copyright (C) 2016-2018, DV Klopfenstein, H Tang, All rights reserved.




"""Download GOA files from the Gene Ontology Annotation (GOA) resource http://www.ebi.ac.uk/GOA."""

__copyright__ = "Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved."
__author__ = "DV Klopfenstein"

import os
import sys
from goatools.base import dnld_file

class DnldGoa(object):
    """Download files from the Gene Ontology Annotation (GOA) resource http://www.ebi.ac.uk/GOA."""

    # European Bioinformatics Institute (EMBL-EBI) ftp site
    ftp_pub = 'ftp://ftp.ebi.ac.uk/pub/'

    # Species available from ftp://ftp.ebi.ac.uk/pub/databases/GO/goa/
    species = [
        'arabidopsis',
        'chicken',
        'cow',
        'dicty',
        'dog',
        'fly',
        'human',
        'mouse',
        #'pdb',
        'pig',
        'rat',
        'uniprot',
        'worm',
        'yeast',
        'zebrafish',
    ]

    species_items = ['complex', 'isoform', 'rna']
    exts = ['gaf', 'gpa', 'gpi']

    def __init__(self):
        self.ftp_src_goa = os.path.join(self.ftp_pub, 'databases/GO/goa/')

    def dnld_goa(self, species, ext='gaf', item=None, fileout=None):
        """Download GOA source file name on EMBL-EBI ftp server."""
        basename = self.get_basename(species, ext, item)
        src = os.path.join(self.ftp_src_goa, species.upper(), "{F}.gz".format(F=basename))
        dst = os.path.join(os.getcwd(), basename) if fileout is None else fileout
        dnld_file(src, dst, prt=sys.stdout, loading_bar=None)
        return dst

    def get_basename(self, species, ext='gaf', item=None):
        """Get GOA basename for a specific species. Ex: goa_human.gaf"""
        assert ext in self.exts, " ".join(self.exts)
        if species == 'uniprot':
            species = 'uniprot_all' if item != 'gcrp' else 'uniprot_gcrp'
        if item is None:
            return 'goa_{SPECIES}.{EXT}'.format(SPECIES=species, EXT=ext)
        assert item in self.species_items
        return 'goa_{SPECIES}_{ITEM}.{EXT}'.format(SPECIES=species, ITEM=item, EXT=ext)


# Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved."



from __future__ import absolute_import

#### # make the module mportable
#### from goatools.go_enrichment import *
#### from . import multiple_testing
#### from . import obo_parser

__author__ = ("Haibao Tang", "DV Klopfenstein")
__copyright__ = "Copyright (C) 2009-2018, Haibao Tang, DV Klopfenstein"
__email__ = "tanghaibao@gmail.com"
__license__ = "BSD"
__status__ = "Development"
__version__ = "0.8.2"



"""Utilities for combining namedtuples."""

__copyright__ = "Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved."
__author__ = "DV Klopfenstein"

import collections as cx

def get_dict_w_id2nts(ids, id2nts, flds, dflt_null=""):
    """Return a new dict of namedtuples by combining "dicts" of namedtuples or objects."""
    assert len(ids) == len(set(ids)), "NOT ALL IDs ARE UNIQUE: {IDs}".format(IDs=ids)
    assert len(flds) == len(set(flds)), "DUPLICATE FIELDS: {IDs}".format(IDs=cx.Counter(flds).most_common())
    usr_id_nt = []
    # 1. Instantiate namedtuple object
    ntobj = cx.namedtuple("Nt", " ".join(flds))
    # 2. Fill dict with namedtuple objects for desired ids
    for item_id in ids:
        # 2a. Combine various namedtuples into a single namedtuple
        nts = [id2nt.get(item_id) for id2nt in id2nts]
        vals = _combine_nt_vals(nts, flds, dflt_null)
        usr_id_nt.append((item_id, ntobj._make(vals)))
    return cx.OrderedDict(usr_id_nt)

def get_list_w_id2nts(ids, id2nts, flds, dflt_null=""):
    """Return a new list of namedtuples by combining "dicts" of namedtuples or objects."""
    combined_nt_list = []
    # 1. Instantiate namedtuple object
    ntobj = cx.namedtuple("Nt", " ".join(flds))
    # 2. Fill dict with namedtuple objects for desired ids
    for item_id in ids:
        # 2a. Combine various namedtuples into a single namedtuple
        nts = [id2nt.get(item_id) for id2nt in id2nts]
        vals = _combine_nt_vals(nts, flds, dflt_null)
        combined_nt_list.append(ntobj._make(vals))
    return combined_nt_list

def combine_nt_lists(lists, flds, dflt_null=""):
    """Return a new list of namedtuples by zipping "lists" of namedtuples or objects."""
    combined_nt_list = []
    # Check that all lists are the same length
    lens = [len(lst) for lst in lists]
    assert len(set(lens)) == 1, \
        "LIST LENGTHS MUST BE EQUAL: {Ls}".format(Ls=" ".join(str(l) for l in lens))
    # 1. Instantiate namedtuple object
    ntobj = cx.namedtuple("Nt", " ".join(flds))
    # 2. Loop through zipped list
    for lst0_lstn in zip(*lists):
        # 2a. Combine various namedtuples into a single namedtuple
        combined_nt_list.append(ntobj._make(_combine_nt_vals(lst0_lstn, flds, dflt_null)))
    return combined_nt_list

def wr_py_nts(fout_py, nts, docstring=None, varname="nts"):
    """Save namedtuples into a Python module."""
    import sys
    import datetime
    if nts:
        with open(fout_py, 'w') as prt:
            first_nt = nts[0]
            nt_name = type(first_nt).__name__
            if docstring is not None:
                prt.write('"""{DOCSTRING}"""\n\n'.format(DOCSTRING=docstring))
            prt.write("# Created: {DATE}\n".format(DATE=str(datetime.date.today())))
            prt.write("import collections as cx\n\n")
            prt.write("nt_fields = [\n")
            for fld in first_nt._fields:
                prt.write('    "{F}",\n'.format(F=fld))
            prt.write("]\n\n")
            prt.write('{NtName} = cx.namedtuple("{NtName}", " ".join(nt_fields))\n\n'.format(
                NtName=nt_name))
            prt.write("# {N:,} items\n".format(N=len(nts)))
            prt.write("{VARNAME} = [\n".format(VARNAME=varname))
            for ntup in nts:
                prt.write("    {NT},\n".format(NT=ntup))
            prt.write("]\n")
            sys.stdout.write(" {N:7,} items WROTE: {PY}\n".format(N=len(nts), PY=fout_py))

def get_unique_fields(fld_lists):
    """Get unique namedtuple fields, despite potential duplicates in lists of fields."""
    flds = []
    fld_set = set([f for flst in fld_lists for f in flst])
    fld_seen = set()
    # Add unique fields to list of fields in order that they appear
    for fld_list in fld_lists:
        for fld in fld_list:
            # Add fields if the field has not yet been seen
            if fld not in fld_seen:
                flds.append(fld)
                fld_seen.add(fld)
    assert len(flds) == len(fld_set)
    return flds

# -- Internal methods ----------------------------------------------------------------
def _combine_nt_vals(lst0_lstn, flds, dflt_null):
    """Given a list of lists of nts, return a single namedtuple."""
    vals = []
    for fld in flds:
        fld_seen = False
        # Set field value using the **first** value seen in list of nt lists(lst0_lstn)
        for nt_curr in lst0_lstn:
            if hasattr(nt_curr, fld):
                vals.append(getattr(nt_curr, fld))
                fld_seen = True
                break
        # Set default value if GO ID or nt value is not present
        if fld_seen is False:
            vals.append(dflt_null)
    return vals

# Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved.



"""Run GOATOOLS Gene Ontology Analysis on Nature 3102 data. Return results."""

__copyright__ = "Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved."
__author__ = "DV Klopfenstein"

import os
import xlrd
from goatools.test_data.genes_NCBI_10090_ProteinCoding import GeneID2nt as GeneID2nt_mus
from goatools.base import get_godag
from goatools.associations import get_assoc_ncbi_taxids
from goatools.go_enrichment import GOEnrichmentStudy

def get_goea_results(keep_if=None):
    """Demonstrate printing a subset of all available fields using two methods."""
    if keep_if is None:
        keep_if = lambda nt: getattr(nt, "p_fdr_bh") < 0.05 # keep if results are significant
    # 1. Gene Ontology Enrichment Analysis
    #    1a. Initialize: Load ontologies, associations, and population gene IDs
    taxid = 10090 # Mouse study
    geneids_pop = GeneID2nt_mus.keys() # Mouse protein-coding genes
    goeaobj = get_goeaobj("fdr_bh", geneids_pop, taxid)
    #    1b. Run GOEA
    geneids_study = get_geneid2symbol("nbt.3102-S4_GeneIDs.xlsx")
    return {
        'goea_results' : goeaobj.run_study(geneids_study, keep_if=keep_if),
        'goeaobj' : goeaobj,
        'geneids_study' : geneids_study,
        'geneids_pop' : geneids_pop,
        'obo_dag':goeaobj.obo_dag}

def get_geneid2symbol(fin_xlsx):
    """Read xlsx file return dictionary with Entrez GeneID keys to Symbol data."""
    gene2symbol = {}
    data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "nbt_3102")
    tbl_genes = "{DIR}/{FIN}".format(DIR=data_dir, FIN=fin_xlsx)
    book = xlrd.open_workbook(tbl_genes)
    sheet = book.sheet_by_index(0)
    for rownum in range(sheet.nrows):
        symbol, geneid, _ = [sheet.cell_value(rownum, c) for c in range(sheet.ncols)]
        if geneid:
            gene2symbol[int(geneid)] = symbol
    return gene2symbol

def get_goeaobj(method, geneids_pop, taxid):
    """Load: ontologies, associations, and population geneids."""
    fin_obo = os.path.join(os.getcwd(), "go-basic.obo")
    godag = get_godag(fin_obo, loading_bar=None)
    assoc_geneid2gos = get_assoc_ncbi_taxids([taxid], loading_bar=None)
    goeaobj = GOEnrichmentStudy(
        geneids_pop,
        assoc_geneid2gos,
        godag,
        propagate_counts=False,
        alpha=0.05,
        methods=[method])
     # godag is also found in goeaobj.godag
    return goeaobj


# Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved.



"""Manage optional GO-DAG attributes."""

__copyright__ = "Copyright (C) 2015-2018, DV Klopfenstein, H Tang, All rights reserved."
__author__ = "DV Klopfenstein"

import re
import collections as cx


# pylint: disable=too-few-public-methods
class OboOptionalAttrs(object):
    """Manage optional GO-DAG attributes."""

    def __init__(self, optional_attrs=None):
        self.attr2cmp = self._init_compile_patterns(optional_attrs)
        self.optional_attrs = optional_attrs

    def update_rec(self, rec, line):
        """Update current GOTerm with optional record."""
        if 'def' in self.optional_attrs and line[:5] == "def: ":
            self._optattr_def(rec, line[5:])
        elif 'synonym' in self.optional_attrs and line[:9] == "synonym: ":
            self._optattr_synonym(rec, line[9:])
        elif 'relationship' in self.optional_attrs and line[:14] == "relationship: ":
            # relationships are now stored in a dict of sets. This mirrors
            # the structure implied by the GO DAG itself. The structure
            # that stores the relationships now looks likes this:
            #
            #  relationship = {
            #     'part_of': set(['GO:0021513', 'GO:0006310']),
            #     'negatively_regulates': set(['GO:0021910'])i
            # }
            self._optattr_relationship(rec, line[14:])
        elif 'xref' in self.optional_attrs and line[:6] == "xref: ":
            self._optattr_xref(rec, line[6:])
        elif 'subset' in self.optional_attrs and line[:8] == "subset: ":
            self._optattr_subset(rec, line[8:])
        elif 'comment' in self.optional_attrs and line[:9] == "comment: ":
            self._optattr_comment(rec, line[9:])

    @staticmethod
    def _optattr_def(rec, value):
        """Store optional attribute, 'defn' in a string."""
        # 'def' is a reserved word in python, do not use it as a Class attr.
        if not hasattr(rec, 'defn'):
            rec.defn = value
        else:
            raise Exception("ATTR(defn) ALREADY SET({VAL})".format(VAL=rec.defn))

    def _optattr_synonym(self, rec, value):
        """Store optional attribute, 'synonym' in a list."""
        if hasattr(rec, 'synonym'):
            rec.synonym.append(self._get_synonym(value))
        else:
            rec.synonym = [self._get_synonym(value)]

    def _optattr_relationship(self, rec, value):
        """Store optional attribute, 'relationship' in a dict of sets."""
        if hasattr(rec, 'relationship'):
            self._add_to_relationship(rec, value)
        else:
            rel, goid = value.split()[:2]
            rec.relationship = {rel: set([goid])}

    def _optattr_xref(self, rec, value):
        """Store optional attribute, 'xref' in a set."""
        if hasattr(rec, 'xref'):
            rec.xref.add(self._get_xref(value))
        else:
            rec.xref = set([self._get_xref(value)])

    @staticmethod
    def _optattr_subset(rec, value):
        """Store optional attribute, 'subset' in a set."""
        if hasattr(rec, 'subset'):
            rec.subset.add(value)
        else:
            rec.subset = set([value])

    @staticmethod
    def _optattr_comment(rec, value):
        """Store optional attribute, 'comment' in a string."""
        # 'def' is a reserved word in python, do not use it as a Class attr.
        if not hasattr(rec, 'comment'):
            rec.comment = value
        else:
            raise Exception("ATTR(comment) ALREADY SET({VAL})".format(VAL=rec.comment))

    def _get_synonym(self, line):
        """Given line, return optional attribute synonym value in a namedtuple."""
        # Example synonyms:
        # "peptidase inhibitor complex" EXACT [GOC:bf, GOC:pr]
        # "regulation of postsynaptic cytosolic calcium levels" EXACT syngo_official_label []
        # "tocopherol 13-hydroxylase activity" EXACT systematic_synonym []
        mtch = self.attr2cmp['synonym'].match(line)
        text, scope, typename, dbxrefs, _ = mtch.groups()
        typename = typename.strip()
        dbxrefs = set(dbxrefs.split(', ')) if dbxrefs else set()
        return self.attr2cmp['synonym nt']._make([text, scope, typename, dbxrefs])

    def _get_xref(self, line):
        """Given line, return optional attribute xref value in a dict of sets."""
        # Ex: xref      Wikipedia:Zygotene
        # Ex: Reactome:REACT_22295 "Addition of a third mannose to ..."
        mtch = self.attr2cmp['xref'].match(line)
        return mtch.group(1).replace(' ', '')

    @staticmethod
    def _add_to_relationship(rec, rel_value):
        """Add GO ID and its relationship to the optional 'relationship' data member."""
        rel, goid = rel_value.split()[:2]
        if rel in rec.relationship:
            rec.relationship[rel].add(goid)
        else:
            rec.relationship[rel] = set([goid])

    @staticmethod
    def _init_compile_patterns(optional_attrs):
        """Compile search patterns for optional attributes if needed."""
        attr2cmp = {}
        if optional_attrs is None:
            return attr2cmp
        # "peptidase inhibitor complex" EXACT [GOC:bf, GOC:pr]
        # "blood vessel formation from pre-existing blood vessels" EXACT systematic_synonym []
        # "mitochondrial inheritance" EXACT []
        # "tricarboxylate transport protein" RELATED [] {comment="WIkipedia:Mitochondrial_carrier"}
        if 'synonym' in optional_attrs:
            attr2cmp['synonym'] = re.compile(r'"(\S.*\S)" ([A-Z]+) (.*)\[(.*)\](.*)$')
            attr2cmp['synonym nt'] = cx.namedtuple("synonym", "text scope typename dbxrefs")
        # Wikipedia:Zygotene
        # Reactome:REACT_27267 "DHAP from Ery4P and PEP, Mycobacterium tuberculosis"
        if 'xref' in optional_attrs:
            attr2cmp['xref'] = re.compile(r'^(\S+:\s*\S+)\b(.*)$')
        return attr2cmp

    @staticmethod
    def get_optional_attrs(optional_attrs):
        """Prepare to store data from user-desired optional fields.

          Not loading these optional fields by default saves in space and speed.
          But allow the possibility for saving these fields, if the user desires,
            Including:
              comment consider def is_class_level is_metadata_tag is_transitive
              relationship replaced_by subset synonym transitive_over xref
        """
        attrs_opt = set(['def', 'defn', 'synonym', 'relationship', 'xref', 'subset', 'comment'])
        # Required attributes are always loaded. All others are optionally loaded.
        # Allow user to specify either: 'def' or 'defn'
        #   'def' is an obo field name, but 'defn' is legal Python attribute name
        getnm = lambda aopt: aopt if aopt != "defn" else "def"
        # pylint: disable=redefined-variable-type
        opts = None
        if isinstance(optional_attrs, str) and optional_attrs in attrs_opt:
            opts = set([getnm(optional_attrs)])
        else:
            opts = set([getnm(f) for f in optional_attrs if f in attrs_opt])
        if opts:
            return opts


# Copyright (C) 2015-2018, DV Klopfenstein, H Tang, All rights reserved.



"""Read a GO Association File (GAF) and store the data in a Python object.

    Annotations available from the Gene Ontology Consortium:
        http://geneontology.org/page/download-annotations

    GAF format:
        http://geneontology.org/page/go-annotation-file-formats
"""

import sys
import os
import re
import collections as cx
from goatools.evidence_codes import EvidenceCodes

__copyright__ = "Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved."
__author__ = "DV Klopfenstein"


# pylint: disable=broad-except,too-few-public-methods,line-too-long
class GafReader(object):
    """Reads a Gene Annotation File (GAF). Returns a Python object."""

    exp_kwdct = set(['allow_missing_symbol'])

    def __init__(self, filename=None, hdr_only=False, prt=sys.stdout, **kws):
        # kws: allow_missing_symbol
        self.kws = {k:v for k, v in kws.items() if k in self.exp_kwdct}
        self.filename = filename
        self.evobj = EvidenceCodes()
        # Initialize associations and header information
        self.hdr = None
        self.associations = self.read_gaf(filename, hdr_only, prt) if filename is not None else []

    def read_gaf(self, fin_gaf, hdr_only, prt):
        """Read GAF file. Store annotation data in a list of namedtuples."""
        nts = []
        ver = None
        hdrobj = GafHdr()
        datobj = None
        lnum = line = -1
        try:
            with open(fin_gaf) as ifstrm:
                for lnum, line in enumerate(ifstrm, 1):
                    # Read header
                    if datobj is None:
                        if line[0] == '!':
                            if ver is None and line[1:13] == 'gaf-version:':
                                ver = line[13:].strip()
                            hdrobj.chkaddhdr(line)
                        else:
                            self.hdr = hdrobj.get_hdr()
                            if hdr_only:
                                return nts
                            datobj = GafData(ver, **self.kws)
                    # Read data
                    if datobj is not None and line[0] != '!':
                        ntgaf = datobj.get_ntgaf(line)
                        if ntgaf is not None:
                            nts.append(ntgaf)
                        else:
                            self.prt_ignore_line(fin_gaf, line, lnum)
        except Exception as inst:
            import traceback
            traceback.print_exc()
            sys.stderr.write("\n  **FATAL in read_gaf: {MSG}\n\n".format(MSG=str(inst)))
            sys.stderr.write("**FATAL: {FIN}[{LNUM}]:\n{L}".format(FIN=fin_gaf, L=line, LNUM=lnum))
            sys.exit(1)
        # GAF file has been read
        if prt is not None:
            prt.write("  READ {N:,} associations: {FIN}\n".format(N=len(nts), FIN=fin_gaf))
        return self.evobj.sort_nts(nts, 'Evidence_Code')

    def prt_summary_anno2ev(self, prt=sys.stdout):
        """Print annotation/evidence code summary."""
        ctr = cx.Counter()
        for ntgaf in self.associations:
            evidence_code = ntgaf.Evidence_Code
            if 'NOT' not in ntgaf.Qualifier:
                ctr[evidence_code] += 1
            elif 'NOT' in ntgaf.Qualifier:
                ctr["NOT {EV}".format(EV=ntgaf.Evidence_Code)] += 1
            else:
                raise Exception("UNEXPECTED INFO")
        self.evobj.prt_ev_cnts(ctr, prt)

    @staticmethod
    def prt_ignore_line(fin_gaf, line, lnum):
        """Print a message saying that we are ignoring an association line."""
        sys.stderr.write("**WARNING: IGNORED {FIN}[{LNUM}]:\n{L}\n\n".format(
            FIN=os.path.basename(fin_gaf), L=line, LNUM=lnum))

class GafData(object):
    """Extracts GAF fields from a GAF line."""

    spec_req1 = [0, 1, 2, 4, 6, 8, 11, 13, 14]

    gafhdr = [ #           Col Req?     Cardinality    Example
        #                  --- -------- -------------- -----------------
        'DB',             #  0 required 1              UniProtKB
        'DB_ID',          #  1 required 1              P12345
        'DB_Symbol',      #  2 required 1              PHO3
        'Qualifier',      #  3 optional 0 or greater   NOT
        'GO_ID',          #  4 required 1              GO:0003993
        'DB_Reference',   #  5 required 1 or greater   PMID:2676709
        'Evidence_Code',  #  6 required 1              IMP
        'With_From',      #  7 optional 0 or greater   GO:0000346
        'Aspect',         #  8 required 1              F
        'DB_Name',        #  9 optional 0 or 1         Toll-like receptor 4
        'DB_Synonym',     # 10 optional 0 or greater   hToll|Tollbooth
        'DB_Type',        # 11 required 1              protein
        'Taxon',          # 12 required 1 or 2         taxon:9606
        'Date',           # 13 required 1              20090118
        'Assigned_By',    # 14 required 1              SGD
    ]

    #                            Col Required Cardinality  Example
    gafhdr2 = [ #                --- -------- ------------ -------------------
        'Annotation_Extension', # 15 optional 0 or greater part_of(CL:0000576)
        'Gene_Product_Form_ID', # 16 optional 0 or 1       UniProtKB:P12345-2
    ]

    gaf_columns = {
        "2.1" : gafhdr + gafhdr2, # !gaf-version: 2.1
        "2.0" : gafhdr + gafhdr2, # !gaf-version: 2.0
        "1.0" : gafhdr}           # !gaf-version: 1.0

    # Expected numbers of columns for various versions
    gaf_numcol = {
        "2.1" : 17,
        "2.0" : 17,
        "1.0" : 15}

    # Expected values for a Qualifier
    exp_qualifiers = set(['not', 'contributes_to', 'colocalizes_with'])

    def __init__(self, ver, allow_missing_symbol=False):
        self.ver = ver
        self.ntgafobj = cx.namedtuple("ntgafobj", " ".join(self.gaf_columns[ver]))
        self.exp_ncol = self.gaf_numcol[ver]
        self.req1 = self.spec_req1 if not allow_missing_symbol else [i for i in self.spec_req1 if i != 2]

    def get_ntgaf(self, line):
        """Return namedtuple filled with data."""
        flds = self._split_line(line)
        return self._get_ntgaf(flds)

    def _split_line(self, line):
        """Split line into field values."""
        line = line.rstrip('\r\n')
        flds = re.split('\t', line)
        assert len(flds) == self.exp_ncol, "EXP {E} COLUMNS, SAW {A}".format(
            A=len(flds), E=self.exp_ncol)
        return flds

    def _get_ntgaf(self, flds):
        """Convert fields from string to preferred format for GAF ver 2.1 and 2.0."""
        # Cardinality
        is_set = False
        is_list = True
        qualifiers = [t.lower() for t in self._rd_fld_vals("Qualifier", flds[3], is_set)]
        db_reference = self._rd_fld_vals("DB_Reference", flds[5], is_set, 1)
        with_from = self._rd_fld_vals("With_From", flds[7], is_set)
        db_name = self._rd_fld_vals("DB_Name", flds[9], is_set, 0)  # , 1)
        db_synonym = self._rd_fld_vals("DB_Synonym", flds[10], is_set)
        taxons = self._rd_fld_vals("Taxon", flds[12], is_list, 1, 2)
        if not self._chk_qty_eq_1(flds):
            return None
        # Additional Formatting
        taxons = self._do_taxons(taxons)
        self._chk_qualifier(qualifiers)
        # Create list of values
        gafvals = [
            flds[0],      # 0  DB
            flds[1],      # 1  DB_ID
            flds[2],      # 2  DB_Symbol
            qualifiers,   # 3  Qualifier
            flds[4],      # 4  GO_ID
            db_reference, # 5  DB_Reference
            flds[6],      # 6  Evidence_Code
            with_from,    # 7  With_From
            flds[8],      # 8  Aspect
            db_name,      # 9  DB_Name
            db_synonym,   # 10 DB_Synonym
            flds[11],     # 11 DB_Type
            taxons,       # 12 Taxon
            flds[12],     # 13 Date
            flds[13]]     # 14 Assigned_By
        # Version 2.x has these additional fields not found in v1.0
        if self.ver[0] == '2':
            gafvals += [
                self._rd_fld_vals("Annotation_Extension", flds[15], is_set),
                self._rd_fld_vals("Gene_Product_Form_ID", flds[16], is_set)]
        return self.ntgafobj._make(gafvals)

    @staticmethod
    def _rd_fld_vals(name, val, set_list_ft=True, qty_min=0, qty_max=None):
        """Further split a GAF value within a single field."""
        if not val and qty_min == 0:
            return [] if set_list_ft else set()
        vals = val.split('|') # Use a pipe to separate entries
        num_vals = len(vals)
        assert num_vals >= qty_min, \
            "FIELD({F}): MIN QUANTITY({Q}) WASN'T MET: {V}".format(F=name, Q=qty_min, V=vals)
        if qty_max is not None:
            assert num_vals <= qty_max, \
                "FIELD({F}): MAX QUANTITY({Q}) EXCEEDED: {V}".format(F=name, Q=qty_max, V=vals)
        return vals if set_list_ft else set(vals)

    def _chk_qualifier(self, qualifiers):
        """Check that qualifiers are expected values."""
        # http://geneontology.org/page/go-annotation-conventions#qual
        for qual in qualifiers:
            assert qual in self.exp_qualifiers, "UNEXPECTED QUALIFIER({Q})".format(Q=qual)

    def _chk_qty_eq_1(self, flds):
        """Check that these fields have only one value: required 1."""
        for col in self.req1:
            if not flds[col]:
                sys.stderr.write("**ERROR: UNEXPECTED REQUIRED VAL({V}) FOR COL({R}):{H}: ".format(
                    V=flds[col], H=self.gafhdr[col], R=col))
                sys.stderr.write("{H0}({DB}) {H1}({ID})\n".format(
                    H0=self.gafhdr[0], DB=flds[0], H1=self.gafhdr[1], ID=flds[1]))
                return False  # Check failed
        return True  # Check passed

    @staticmethod
    def _do_taxons(taxons):
        """Taxon"""
        taxons = [int(v[6:]) for v in taxons] # strip "taxon:"
        num_taxons = len(taxons)
        assert num_taxons == 1 or num_taxons == 2
        return taxons


class GafHdr(object):
    """Used to build a GAF header."""

    cmpline = re.compile(r'^!(\w[\w\s-]+:.*)$')

    def __init__(self):
        self.gafhdr = []

    def get_hdr(self):
        """Return GAF header data as a string paragragh."""
        return "\n".join(self.gafhdr)

    def chkaddhdr(self, line):
        """If this line contains desired header info, save it."""
        mtch = self.cmpline.search(line)
        if mtch:
            self.gafhdr.append(mtch.group(1))

# Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved."



"""Report the level/depth summaries of all GO terms or a subset of GO terms.

   Level is the length of a shortest path to a GO term.
   Depth is the length of a longest path to a GO term.

    Example:

    >>> godag = GODag("go-basic.obo")
    >>> reporter = RptLevDepth(godag)
    >>> reporter.write_summary_cnts_all()
        go-basic.obo: format-version(1.2) data-version(releases/2016-03-01)
        46162 nodes imported
        Dep <-Depth Counts->  <-Level Counts->
        Lev   BP    MF    CC    BP    MF    CC
        --- ----  ----  ----  ----  ----  ----
        00     1     1     1     1     1     1
        01    24    19    24    24    19    24
        02   126   131   195   222   154   336
        03   955   493   502  1913   739  1145
        04  1950  1464   566  4506  1812  1300
        05  3371  3765   994  6958  3985   763
        06  4275  1784   748  6945  1888   268
        07  4634  1005   549  4890   903    53
        08  4125   577   216  2022   352     6
        09  3478   315    86   739   109     1
        10  2364   164    12   188    38     0
        11  1583   170     3    37    21     0
        12  1014    68     1     1     0     0
        13   400    49     0     0     0     0
        14   107    13     0     0     0     0
        15    28     3     0     0     0     0
        16    11     0     0     0     0     0

"""

__copyright__ = "Copyright (C) 2015-2018, DV Klopfenstein, H Tang, All rights reserved."
__author__ = "DV Klopfenstein"

import sys
import collections as cx

def prt_lev_depth(godag, prt=sys.stdout):
    """Write the counts of GO terms found at each level and depth of the GO DAG."""
    RptLevDepth(godag, prt).write_summary_cnts_all()

class RptLevDepth(object):
    """Reports a summary of GO term levels in depths."""

    nss = ['biological_process', 'molecular_function', 'cellular_component']

    def __init__(self, obodag, log=sys.stdout):
        self.obo = obodag
        self.log = log
        self.title = "GO Counts in {VER}".format(VER=self.obo.version)

    def wr_xlsx(self, fout_xlsx):
        """Write counts of GO terms at all levels and depths."""
        from goatools.wr_tbl import wr_xlsx
        kws = {
            'title' : self.title(),
            'hdrs' :["Dep/Lev",
                     "BP Depth", "MF Depth", "CC Depth",
                     "BP Level", "MF Level", "CC Level"]}
        wr_xlsx(fout_xlsx, self.get_data(), **kws)

    def wr_txt(self, fout_txt):
        """Write counts of GO terms at all levels and depths."""
        from goatools.wr_tbl import prt_txt
        data = self.get_data()
        with open(fout_txt, 'w') as prt:
            prtfmt = "{Depth_Level:>7} " \
                     "{BP_D:6,} {MF_D:6,} {CC_D:>6,} " \
                     "{BP_L:>6,} {MF_L:>6,} {CC_L:>6,}\n"
            prt.write("{TITLE}\n\n".format(TITLE=self.title))
            prt.write("         |<---- Depth ---->|  |<---- Level ---->|\n")
            prt.write("Dep/Lev     BP     MF     CC     BP     MF     CC\n")
            prt.write("-------  -----  -----  -----  -----  -----  -----\n")
            prt_txt(prt, data, prtfmt=prtfmt, title=self.title)
            sys.stdout.write("  {N:>5,} items WROTE: {TXT}\n".format(
                N=len(data), TXT=fout_txt))

    def prttex_summary_cnts_all(self, prt=sys.stdout):
        cnts = self.get_cnts_levels_depths_recs(set(self.obo.values()))
        self._prttex_summary_cnts(prt, cnts)

    def write_summary_cnts_all(self):
        """Write summary of level and depth counts for all active GO Terms."""
        cnts = self.get_cnts_levels_depths_recs(set(self.obo.values()))
        self._write_summary_cnts(cnts)

    def write_summary_cnts(self, go_ids):
        """Write summary of level and depth counts for specific GO ids."""
        obo = self.obo
        cnts = self.get_cnts_levels_depths_recs([obo.get(GO) for GO in go_ids])
        self._write_summary_cnts(cnts)

    def write_summary_cnts_goobjs(self, goobjs):
        """Write summary of level and depth counts for active GO Terms."""
        cnts = self.get_cnts_levels_depths_recs(goobjs)
        self._write_summary_cnts(cnts)

    def _prttex_summary_cnts(self, prt, cnts):
        """Write summary of level and depth counts for active GO Terms."""
        # Count level(shortest path to root) and depth(longest path to root)
        # values for all unique GO Terms.
        prt.write("\n\n% LaTeX Table for GO counts at each level and depth in the GO DAG\n")
        prt.write("\\begin{table}[bt]\n")
        prt.write(" ".join(["\\caption{\label{tab:godepcnt}", self.title, "}\n"]))
        prt.write("\\begin{tabular}{r r r r r r r}\n")
        prt.write("\\toprule\n")
        prt.write("Depth/Level & "
                  "BP Depth & MF Depth & CC Depth & "
                  "BP Level & MF Level & CC Level \\\\\n")
        prt.write("\midrule\n")
        max_val = max(max(dep for dep in cnts['depth']), max(lev for lev in cnts['level']))
        for i in range(max_val+1):
            vals = ['{:>5}'.format(cnts[desc][i][ns]) for desc in cnts for ns in self.nss]
            self.log.write('{:>02} & {} \\\\\n'.format(i, ' & '.join(vals)))
        prt.write("\\bottomrule\n")
        prt.write("\\end{tabular}\n")
        prt.write("\\end{table}\n")

    def _write_summary_cnts(self, cnts):
        """Write summary of level and depth counts for active GO Terms."""
        # Count level(shortest path to root) and depth(longest path to root)
        # values for all unique GO Terms.
        max_val = max(max(dep for dep in cnts['depth']),
                      max(lev for lev in cnts['level']))
        self.log.write('Dep <-Depth Counts->  <-Level Counts->\n')
        self.log.write('Lev   BP    MF    CC    BP    MF    CC\n')
        self.log.write('--- ----  ----  ----  ----  ----  ----\n')
        for i in range(max_val+1):
            vals = ['{:>5}'.format(cnts[desc][i][ns]) for desc in cnts for ns in self.nss]
            self.log.write('{:>02} {}\n'.format(i, ' '.join(vals)))

    @staticmethod
    def get_cnts_levels_depths_recs(recs):
        """Collect counts of levels and depths in a Group of GO Terms."""
        cnts = cx.defaultdict(lambda: cx.defaultdict(cx.Counter))
        for rec in recs:
            if rec is not None and not rec.is_obsolete:
                cnts['level'][rec.level][rec.namespace] += 1
                cnts['depth'][rec.depth][rec.namespace] += 1
        return cnts

    def get_data(self):
        """Collect counts of GO terms at all levels and depths."""
        # Count level(shortest path to root) and depth(longest path to root)
        # values for all unique GO Terms.
        data = []
        ntobj = cx.namedtuple("NtGoCnt", "Depth_Level BP_D MF_D CC_D BP_L MF_L CC_L")
        cnts = self.get_cnts_levels_depths_recs(set(self.obo.values()))
        max_val = max(max(dep for dep in cnts['depth']), max(lev for lev in cnts['level']))
        for i in range(max_val+1):
            vals = [i] + [cnts[desc][i][ns] for desc in cnts for ns in self.nss]
            data.append(ntobj._make(vals))
        return data

# Copyright (C) 2015-2018, DV Klopfenstein, H Tang, All rights reserved."



"""Given GO ids and an obo, creates a small sub-graph DAG.

   Sub-graphs can be used to create shortcut paths and eliminate nodes.
"""

from collections import defaultdict
from goatools.godag_small import GODagSmall

__copyright__ = "Copyright (C) 2016-2018, DV Klopfenstein, H Tang, All rights reserved."
__author__ = "DV Klopfenstein"

class OboToGoDagSmall(object):
    """Given a list of GO ids and the full obo dag, return a sub GO DAG graph."""

    def __init__(self, **kws):
        # Initial setup
        self.godag = GODagSmall()
        self.godag.go_sources = None
        self._init_go2obj(**kws)
        # Initialize remainder of graph
        self.traverse_child = kws['traverse_child'] if 'traverse_child' in kws else False
        self.traverse_parent = kws['traverse_parent'] if 'traverse_parent' in kws else True
        self.seen_cids = set() if self.traverse_parent else None
        self.seen_pids = set() if self.traverse_child else None
        self._init()
        assert self.traverse_child or self.traverse_parent, "NO EDGES IN GRAPH"

    def _init_go2obj(self, **kws):
        """Initialize go2obj in small dag for source gos."""
        if 'goids' in kws and 'obodag' in kws:
            self.godag.go_sources = kws['goids']
            obo = kws['obodag']
            for goid in self.godag.go_sources:
                self.godag.go2obj[goid] = obo[goid]
        elif 'goid2goobj' in kws:
            goid2goobj = kws['goid2goobj']
            self.godag.go_sources = goid2goobj.keys()
            for goid, goobj in goid2goobj.items():
                self.godag.go2obj[goid] = goobj
        elif 'goea_results' in kws:
            goea_results = kws['goea_results']
            self.godag.go_sources = [rec.GO for rec in goea_results]
            self.godag.go2obj = {rec.GO:rec.goterm for rec in goea_results}

    def _init(self):
        """Given GO ids and GOTerm objects, create mini GO dag."""
        for goid in self.godag.go_sources:
            goobj = self.godag.go2obj[goid]
            self.godag.go2obj[goid] = goobj
            # Traverse up parents
            if self.traverse_parent and goid not in self.seen_cids:
                self._traverse_parent_objs(goobj)
            # Traverse down children
            if self.traverse_child and goid not in self.seen_pids:
                self._traverse_child_objs(goobj)

    def _traverse_parent_objs(self, goobj_child):
        """Traverse from source GO up parents."""
        child_id = goobj_child.id
        # mark child as seen
        self.seen_cids.add(child_id)
        self.godag.go2obj[child_id] = goobj_child
        # Loop through parents of child object
        for parent_obj in goobj_child.parents:
            parent_id = parent_obj.id
            self.godag.p_from_cs[parent_id].add(child_id)
            # If parent has not been seen, traverse
            if parent_id not in self.seen_cids:
                self._traverse_parent_objs(parent_obj)
            
    def _traverse_child_objs(self, goobj_parent):
        """Traverse from source GO down children."""
        parent_id = goobj_parent.id
        # mark parent as seen
        self.seen_pids.add(parent_id)
        self.godag.go2obj[parent_id] = goobj_parent
        # Loop through children
        for child_obj in goobj_parent.children:
            child_id = child_obj.id
            self.godag.p2cs[parent_id].add(child_id)
            # If child has not been seen
            if child_id not in self.seen_pids:
                self._traverse_child_objs(child_obj)

# Copyright (C) 2016-2018, DV Klopfenstein, H Tang, All rights reserved.




"""Options for calculating uncorrected p-values."""

__copyright__ = "Copyright (C) 2016-2018, DV Klopfenstein, H Tang et al., All rights reserved."
__author__ = "DV Klopfenstein"

import collections as cx
import sys

class PvalCalcBase(object):
    """Base class for initial p-value calculations."""

    def __init__(self, name, pval_fnc, log):
        self.log = log
        self.name = name
        self.pval_fnc = pval_fnc

    def calc_pvalue(self, study_count, study_n, pop_count, pop_n):
        """pvalues are calculated in derived classes."""
        fnc_call = "calc_pvalue({SCNT}, {STOT}, {PCNT} {PTOT})".format(
            SCNT=study_count, STOT=study_n, PCNT=pop_count, PTOT=pop_n)
        raise Exception("NOT IMPLEMENTED: {FNC_CALL} using {FNC}.".format(
            FNC_CALL=fnc_call, FNC=self.pval_fnc))


class FisherClass(PvalCalcBase):
    """From the 'fisher' package, use function, pvalue_population."""

    def __init__(self, name, log):
        import fisher
        super(FisherClass, self).__init__(name, fisher.pvalue_population, log)

    def calc_pvalue(self, study_count, study_n, pop_count, pop_n):
        """Calculate uncorrected p-values."""
        # k, n = study_true, study_tot,
        # K, N = population_true, population_tot
        # def pvalue_population(int k, int n, int K, int N): ...
        return self.pval_fnc(study_count, study_n, pop_count, pop_n).two_tail


class FisherScipyStats(PvalCalcBase):
    """From the scipy stats package, use function, fisher_exact."""

    def __init__(self, name, log):
        from scipy import stats
        super(FisherScipyStats, self).__init__(name, stats.fisher_exact, log)

    def calc_pvalue(self, study_count, study_n, pop_count, pop_n):
        """Calculate uncorrected p-values."""
        # http://docs.scipy.org/doc/scipy-0.17.0/reference/generated/scipy.stats.fisher_exact.html
        #
        #         Atlantic  Indian                              YES       NO
        # whales     8        2    | 10 whales    study_genes    8 scnt   2    | 10 = study_n
        # sharks     1        5    |  6 sharks    not s_genes    1        5    |  6
        #         --------  ------                            --------   -----
        #            9        7      16 = pop_n     pop_genes    9 pcnt   7      16 = pop_n
        #
        # We use the preceeding table to find the p-value for whales/sharks:
        #
        # >>> import scipy.stats as stats
        # >>> oddsratio, pvalue = stats.fisher_exact([[8, 2], [1, 5]])
        #                                              a  b    c  d
        a = study_count
        b = study_n - study_count
        c = pop_count - study_count
        d = pop_n - pop_count - b
        # stats.fisher_exact returns oddsratio, pval_uncorrected
        _, p_uncorrected = self.pval_fnc( [[a, b], [c, d]])
        return p_uncorrected


class FisherFactory(object):
    """Factory for choosing a fisher function."""

    options = cx.OrderedDict([
        ('fisher', FisherClass),
        ('fisher_scipy_stats', FisherScipyStats),
    ])

    def __init__(self, **kws):
        self.log = kws['log'] if 'log' in kws else sys.stdout
        self.pval_fnc_name = kws["pvalcalc"] if "pvalcalc" in kws else "fisher"
        self.pval_obj = self._init_pval_obj()

    def _init_pval_obj(self):
        """Returns a Fisher object based on user-input."""
        if self.pval_fnc_name in self.options.keys():
            try:
                fisher_obj = self.options[self.pval_fnc_name](self.pval_fnc_name, self.log)
            except ImportError:
                print("fisher module not installed.  Falling back on scipy.stats.fisher_exact")
                fisher_obj = self.options['fisher_scipy_stats']('fisher_scipy_stats', self.log)

            return fisher_obj

        raise Exception("PVALUE FUNCTION({FNC}) NOT FOUND".format(FNC=self.pval_fnc_name))

    def __str__(self):
        return " ".join(self.options.keys())


# Copyright (C) 2016-2018, DV Klopfenstein, H Tang et al., All rights reserved.



"""Helper Class and functions for writing tables."""

import sys


class WrXlsxParams(object):
    """Class containing user-specified or default parameters."""

    def __init__(self, **kws):
        self.kws = kws
        self.title = kws.get('title', None)
        self.fld2fmt = kws.get('fld2fmt', None)
        self.fld2col_widths = kws.get('fld2col_widths', None)
        # User may specify to skip rows based on values in row
        self.prt_if = kws.get('prt_if', None)
        # Initialize sort_by
        self.sort_by = kws.get('sort_by', None)


class WrXlsx(object):
    """Class to store/manage Excel spreadsheet parameters."""

    dflt_fmt_hdr = {'top':0, 'bottom':0, 'left':0, 'right':0, 'bold':True}

    def __init__(self, fout_xlsx, nt_flds, **kws):
        # KEYWORDS FOR WRITING DATA:
        self.vars = WrXlsxParams(**kws)
        # Workbook
        from xlsxwriter import Workbook
        self.workbook = Workbook(fout_xlsx)
        self.wbfmtobj = WbFmt(nt_flds, self.workbook, **kws)
        self.hdrs = self.wbfmtobj.get_hdrs(**kws)
        self.fmt_hdr = self.workbook.add_format(self.dflt_fmt_hdr)

    def wr_title(self, worksheet, row_idx=0):
        """Write title (optional)."""
        if self.vars.title is not None:
            return self.wr_row_mergeall(worksheet, self.vars.title, self.fmt_hdr, row_idx)
        return row_idx

    def wr_row_mergeall(self, worksheet, txtstr, fmt, row_idx):
        """Merge all columns and place text string in widened cell."""
        hdridxval = len(self.hdrs) - 1
        worksheet.merge_range(row_idx, 0, row_idx, hdridxval, txtstr, fmt)
        return row_idx + 1

    def wr_hdrs(self, worksheet, row_idx):
        """Print row of column headers"""
        for col_idx, hdr in enumerate(self.hdrs):
            worksheet.write(row_idx, col_idx, hdr, self.fmt_hdr)
        row_idx += 1
        return row_idx

    def wr_data(self, xlsx_data, row_i, worksheet):
        """Write data into xlsx worksheet."""
        fld2fmt = self.vars.fld2fmt
        # User may specify to skip rows based on values in row
        prt_if = self.vars.prt_if
        # User may specify a subset of columns to print or
        # a column ordering different from the _fields seen in the namedtuple
        prt_flds = self.wbfmtobj.get_prt_flds()
        get_wbfmt = self.wbfmtobj.get_wbfmt
        if self.vars.sort_by is not None:
            xlsx_data = sorted(xlsx_data, key=self.vars.sort_by)
        try:
            for data_nt in xlsx_data:
                if prt_if is None or prt_if(data_nt):
                    wbfmt = get_wbfmt(data_nt)
                    # Print an xlsx row by printing each column in order.
                    for col_i, fld in enumerate(prt_flds):
                        try:
                            # If fld "format_txt" present, use val for formatting, but don't print.
                            val = getattr(data_nt, fld, "")
                            # Optional user-formatting of specific fields, eg, pval: "{:8.2e}"
                            # If field value is empty (""), don't use fld2fmt
                            if fld2fmt is not None and fld in fld2fmt and val != "" and val != "*":
                                val = fld2fmt[fld].format(val)
                            worksheet.write(row_i, col_i, val, wbfmt)
                        except:
                            raise RuntimeError(self._get_err_msg(row_i, col_i, fld, val, prt_flds))
                    row_i += 1
        except RuntimeError as inst:
            import traceback
            traceback.print_exc()
            sys.stderr.write("\n  **FATAL in wr_data: {MSG}\n\n".format(MSG=str(inst)))
            sys.exit(1)
        return row_i

    @staticmethod
    def _get_err_msg(row, col, fld, val, prt_flds):
        """Return an informative message with details of xlsx write attempt."""
        import traceback
        traceback.print_exc()
        err_msg = (
            "ROW({R}) COL({C}) FIELD({F}) VAL({V})\n".format(R=row, C=col, F=fld, V=val),
            "PRINT FIELDS({N}): {F}".format(N=len(prt_flds), F=" ".join(prt_flds)))
        return "\n".join(err_msg)

    def add_worksheet(self):
        """Add a worksheet to the workbook."""
        wsh = self.workbook.add_worksheet()
        if self.vars.fld2col_widths is not None:
            self.set_xlsx_colwidths(wsh, self.vars.fld2col_widths, self.wbfmtobj.get_prt_flds())
        return wsh

    @staticmethod
    def set_xlsx_colwidths(worksheet, fld2col_widths, fldnames):
        """Set xlsx column widths using fld2col_widths."""
        for col_idx, fld in enumerate(fldnames):
            col_width = fld2col_widths.get(fld, None)
            if col_width is not None:
                worksheet.set_column(col_idx, col_idx, col_width)


class WbFmt(object):
    """Manages Workbook format objects."""

    dflt_wbfmtdict = [
        # Colors from "Grey Grey Color Palette" http://www.color-hex.com/color-palette/20477
        # http://www.color-hex.com/color/808080
        # http://www.color-hex.com/color/d3d3d3
        {'border':0},
        {'top':0, 'bottom':0, 'left':0, 'right':0, 'bold':True, 'bg_color':'#eeeded'},
        {'top':0, 'bottom':0, 'left':0, 'right':0, 'bold':True, 'bg_color':'#d3d3d3'},
        {'border':1, 'bold':True}]

        # Save user-provided cell format, if provided. If not, use default.
    def __init__(self, nt_flds, workbook, **kws):
        self.prt_flds = kws.get('prt_flds', nt_flds)
        self.b_format_txt = 'format_txt' in self.prt_flds
        self.ntfld_wbfmt = kws.get('ntfld_wbfmt', None)
        self.fmtname2wbfmtobj = self._init_fmtname2wbfmtobj(workbook, **kws)
        self.b_plain = not self.b_format_txt and self.ntfld_wbfmt is None

    def get_prt_flds(self):
        """Remove namedtuple fields used for formatting rows, but not printed."""
        return [f for f in self.prt_flds if f != "format_txt"]

    def get_hdrs(self, **kws):
        """Initialize column headers."""
        hdrs = get_hdrs(self.prt_flds, **kws)
        # Values in a "format_txt" "column" are used for formatting, not printing
        return [h for h in hdrs if h != "format_txt"]

    def _init_fmtname2wbfmtobj(self, workbook, **kws):
        """Initialize fmtname2wbfmtobj."""
        wbfmtdict = [
            kws.get('format_txt0', self.dflt_wbfmtdict[0]),
            kws.get('format_txt1', self.dflt_wbfmtdict[1]),
            kws.get('format_txt2', self.dflt_wbfmtdict[2]),
            kws.get('format_txt3', self.dflt_wbfmtdict[3])]
        fmtname2wbfmtobj = {
            'plain': workbook.add_format(wbfmtdict[0]),
            'plain bold': workbook.add_format(wbfmtdict[3]),
            'very light grey' : workbook.add_format(wbfmtdict[1]),
            'light grey' :workbook.add_format(wbfmtdict[2])}
        # Use a xlsx namedtuple field value to set row color
        ntval2wbfmtdict = kws.get('ntval2wbfmtdict', None)
        if ntval2wbfmtdict is not None:
            for ntval, wbfmtdict in ntval2wbfmtdict.items():
                fmtname2wbfmtobj[ntval] = workbook.add_format(wbfmtdict)
            if 'ntfld_wbfmt' not in kws:
                sys.stdout.write("**WARNING: 'ntfld_wbfmt' NOT PRESENT\n")
        return fmtname2wbfmtobj

    def get_wbfmt(self, data_nt=None):
        """Return format for text cell."""
        if data_nt is None or self.b_plain:
            return self.fmtname2wbfmtobj.get('plain')
        # User namedtuple field/value for color
        if self.ntfld_wbfmt is not None:
            return self.__get_wbfmt_usrfld(data_nt)
        # namedtuple format_txt for color/bold/border
        if self.b_format_txt:
            wbfmt = self.__get_wbfmt_format_txt(data_nt)
            if wbfmt is not None:
                return wbfmt
        # 'ntfld_wbfmt': namedtuple field which contains a value used as a key for a xlsx format
        # 'ntval2wbfmtdict': namedtuple value and corresponding xlsx format dict. Examples:
        return self.fmtname2wbfmtobj.get('plain')

    def __get_wbfmt_usrfld(self, data_nt):
        """Return format for text cell from namedtuple field specified by 'ntfld_wbfmt'"""
        if self.ntfld_wbfmt is not None:
            ntval = getattr(data_nt, self.ntfld_wbfmt, None) # Ex: 'section'
            if ntval is not None:
                return self.fmtname2wbfmtobj.get(ntval, None)

    def __get_wbfmt_format_txt(self, data_nt):
        """Return format for text cell from namedtuple field, 'format_txt'."""
        format_txt_val = getattr(data_nt, "format_txt")
        if format_txt_val == 1:
            return self.fmtname2wbfmtobj.get("very light grey")
        if format_txt_val == 2:
            return self.fmtname2wbfmtobj.get("light grey")
        return self.fmtname2wbfmtobj.get(format_txt_val)

    def get_fmt_section(self):
        """Grey if printing header GOs and plain if not printing header GOs."""
        if self.b_format_txt:
            return self.fmtname2wbfmtobj.get("light grey")
        return self.fmtname2wbfmtobj.get("plain bold")


def get_hdrs(flds_all, **kws):
    """Return headers, given user-specified key-word args."""
    # Return Headers if the user explicitly lists them.
    hdrs = kws.get('hdrs', None)
    if hdrs is not None:
        return hdrs
    # User may specify a subset of fields or a column order using prt_flds
    if 'prt_flds' in kws:
        return kws['prt_flds']
    # All fields in the namedtuple will be in the headers
    return flds_all

#  Copyright (C) 2015-2018. All rights reserved.



"""Annotation Extension for relational expressions.

   https://link.springer.com/protocol/10.1007/978-1-4939-3743-1_17

"""

__copyright__ = "Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved."
__author__ = "DV Klopfenstein"

# pylint: disable=too-few-public-methods
class AnnotationExtension(object):
    """Annotation Extension for relational expressions."""

    def __init__(self, relation, entity):
        # Relationship between GO term and the entity
        self.relation = relation
        # An identifier for a database object or ontology term
        self.entity = entity

    def __str__(self):
        return "{R}({E})".format(R=self.relation, E=self.entity)


# Copyright (C) 2016-2018, DV Klopfenstein, H Tang. All rights reserved."



# Tools for managing data in The GOA database: Gene Ontology annotation (GOA)
# resource (http://www.ebi.ac.uk/GOA) resource (http://www.ebi.ac.uk/GOA)

# https://academic.oup.com/nar/article/43/D1/D1057/2437623/The-GOA-database-Gene-Ontology-annotation-updates



"""Script for downloading associations from GO Database for a given species."""
# -*- coding: UTF-8 -*-
from __future__ import print_function

"""
Fetch associations for S Pombe (NCBI Taxonomy ID 4896)
>>> python {SCR} --taxon_id 4896 -o spombe.assocs

The output format is the same as for 

    SPBC18H10.18c   GO:0008150;GO:0005737;GO:0016021;GO:0003674
    mis13   GO:0005634;GO:0005515;GO:0031617;GO:0000070;GO:0000779;GO:0000941;GO:0000444
    fap7    GO:0016887;GO:0005634;GO:0000462;GO:0004017;GO:0005524;GO:0017111;GO:0005829
    SPCC594.01      GO:0008150;GO:0005575;GO:0003674

Other useful taxon IDs:

 - 9606 Human
 - 10090 Mouse

Note that sometimes it is necessary to use the strain ID. See: http://geneontology.org/page/download-annotations

TODO: allow use of taxon labels, allow custom filtering

""".format(SCR=__file__)

import pysolr
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))



def main():
    """Fetch simple gene-term assocaitions from Golr using bioentity document type, one line per gene."""
    import argparse
    prs = argparse.ArgumentParser(__doc__,
                                  formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    prs.add_argument('--taxon_id', type=str,
                     help='NCBI taxon ID, must match exact species/strain used by GO Central, e.g. 4896 for S Pombe')
    prs.add_argument('--golr_url', default='http://golr.geneontology.org/solr/', type=str,
                     help='NCBI taxon ID, must match exact species/strain used by GO Central, e.g. 4896 for S Pombe')
    prs.add_argument('-o', default=None, type=str,
                     help="Specifies the name of the output file")
    prs.add_argument('--max_rows', default=100000, type=int,
                     help="maximum rows to be fetched")

    args = prs.parse_args()

    solr = pysolr.Solr(args.golr_url, timeout=30)

    sys.stderr.write("TAX:"+args.taxon_id+"\n")
    results = solr.search(q='document_category:"bioentity" AND taxon:"NCBITaxon:'+args.taxon_id+'"',
                          fl='bioentity_label,annotation_class_list', rows=args.max_rows)
    sys.stderr.write("NUM GENES:"+str(len(results))+"\n")
    if (len(results) ==0):
        sys.stderr.write("NO RESULTS")
        exit(1)
    if (len(results) == args.max_rows):
        sys.stderr.write("max_rows set too low")
        exit(1)

    file_out = sys.stdout if args.o is None else open(args.o, 'w')
    for r in results:
        gene_symbol = r['bioentity_label']
        sys.stderr.write(gene_symbol+"\n")
        if 'annotation_class_list' in r:
            file_out.write(r['bioentity_label']+"\t" + ';'.join(r['annotation_class_list'])+"\n")
        else:
            sys.stderr.write("no annotations for "+gene_symbol+"\n")
            

    if args.o is not None:
        file_out.close()
        sys.stdout.write("  WROTE: {}\n".format(args.o))

if __name__ == "__main__":
    main()



"""Test the loading of the optional GO term fields."""
# https://owlcollab.github.io/oboformat/doc/GO.format.obo-1_4.html

__copyright__ = "Copyright (C) 2010-2018, DV Klopfenstein, H Tang, All rights reserved."
__author__ = "DV Klopfenstein"


import os
import sys
import re
import collections as cx
import timeit
import datetime
from goatools.obo_parser import GODag
from goatools.base import download_go_basic_obo


class OptionalAttrs(object):
    """Holds data for GO relationship test."""

    repo = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../..")
    cmpfld = re.compile(r'^(\S+)\s*:\s*(\S.*\S)\s*$')  # Field line pattern
    exp_scopes = set(['EXACT', 'BROAD', 'NARROW', 'RELATED'])
    exp_xrefpat = re.compile(r'^\S+:\S+$')
    # Required attributes are always loaded
    exp_req = set(['name', 'id', 'is_obsolete', 'namespace', 'alt_ids', 'is_a'])
    # Generated attributes
    exp_gen = set(['level', 'depth', 'parents', 'children', '_parents'])

    attrs_req = set(['id', 'namespace', 'name', 'is_a', 'alt_ids'])
    attrs_scalar = set(['id', 'namespace', 'name', 'def', 'comment'])
    attrs_set = set(['xref', 'subset', 'alt_id'])

    def __init__(self, fin_obo, opt_field=None):
        self.opt = opt_field  # None causes all fields to read to exp dict
        self.obo = os.path.join(self.repo, fin_obo)
        self._init_dnld_dag()
        self.go2obj = {o.id:o for o in self.load_dag(opt_field).values()}
        self.dcts = self._init_go2dct()  # dagdct go2dct typdefdct flds
        self.go2dct = {go:d for go, d in self.dcts['go2dct'].items() if go in self.go2obj}
        self.num_tot = len(self.go2obj)
        self._chk_required()
        self._chk_parents()
        self._set_exp_children()
        self._chk_children()

    def prt_summary(self, prt=sys.stdout):
        """Print percentage of GO IDs that have a specific relationship."""
        prt.write("\nThese fields appear at least once on a GO term:\n")
        # Ex: 28,951 of 44,948 (64%) GO IDs has field(synonym)
        for relname, cnt in self._get_cnts_gte1().most_common():
            self._prt_perc(cnt, relname, prt)
        prt.write("\nMaximum number of fields on a GO term:\n")
        for fld, maxqty in sorted(self._get_cnts_max().items(), key=lambda t: t[1]):
            prt.write("    {MAX:3} {MRK} {FLD}\n".format(
                MAX=maxqty, MRK=self._get_fldmrk(fld), FLD=fld))

    def _chk_parents(self):
        """Check that parent relationships."""
        for goobj in self.go2obj.values():
            exp_dct = self.go2dct[goobj.id]
            if 'is_a' in exp_dct:
                # pylint: disable=protected-access
                exp_parents = exp_dct['is_a']
                act_parents = goobj._parents
                assert exp_parents == act_parents
            else:
                assert not goobj.parents

    def _chk_children(self):
        """Check that parent relationships."""
        for goobj in self.go2obj.values():
            exp_dct = self.go2dct[goobj.id]
            if '_children' in exp_dct:
                exp_children = exp_dct['_children']
                act_children = set(o.id for o in goobj.children)
                assert exp_children == act_children
            else:
                assert not goobj.children

    def _set_exp_children(self):
        """Fill expected child GO IDs."""
        # Initialize empty sets for child GO IDs
        for exp_dct in self.go2dct.values():
            exp_dct['_children'] = set()
        # Loop thru all GO IDs
        for goid_child, exp_dct in self.go2dct.items():
            if 'is_a' in exp_dct:
                # Add current GO ID to all of it's parents' set of children
                for goid_parent in exp_dct['is_a']:
                    self.go2dct[goid_parent]['_children'].add(goid_child)

    def _chk_required(self):
        """Check the required attributes."""
        for goid, goobj in self.go2obj.items():
            godct = self.go2dct[goid]
            assert goobj.id == godct['GO']
            assert goobj.namespace == next(iter(godct['namespace'])), godct
            assert goobj.name == next(iter(godct['name']))
            self._chk_is_obsolete(goobj, godct)
            self._chk_alt_ids(goobj, godct)

    @staticmethod
    def _chk_alt_ids(goobj, godct):
        """Check 'alt_ids' required attribute."""
        if 'alt_id' in godct:
            assert godct['alt_id'] == goobj.alt_ids
        else:
            assert not goobj.alt_ids

    @staticmethod
    def _chk_is_obsolete(goobj, godct):
        """Check 'is_obsolete' required attribute."""
        act_obso = getattr(goobj, 'is_obsolete', None)
        if act_obso:
            assert 'is_obsolete' in godct, "EXP({})\nACT({})".format(
                godct, getattr(goobj, 'is_obsolete', None))
        else:
            assert 'is_obsolete' not in godct, "EXP({})\nACT({})".format(
                godct, getattr(goobj, 'is_obsolete', None))

    def chk_no_optattrs(self):
        """Check that only the optional attributes requested are the attributes implemented."""
        # name is_obsolete namespace id alt_ids
        # level namespace depth parents children _parents
        exp_flds = self.exp_req.union(self.exp_gen)
        for goobj in self.go2obj.values():
            assert not set(vars(goobj).keys()).difference(exp_flds)
            # print(vars(goobj).keys())
            # print(" ".join(vars(goobj).keys()))

    def chk_xref(self, prt=None):
        """Check synonyms."""
        # Get GO IDs which are expected to have synonyms
        goids = set(go for go, d in self.go2dct.items() if 'xref' in d)
        for goid in goids:
            goobj = self.go2obj[goid]
            xrefs = getattr(goobj, 'xref', None)
            assert xrefs is not None, "{GO} MISSING XREF".format(GO=goid)
            # Iterate through list of synonym data stored in named tuples
            for dbxref in xrefs:
                if prt is not None:
                    prt.write("{GO} {DBXREF}\n".format(GO=goid, DBXREF=dbxref))
                assert self.exp_xrefpat.match(dbxref), "INVALID XREF FORMAT"

    def chk_synonyms(self, prt=None):
        """Check synonyms."""
        # Get GO IDs which are expected to have synonyms
        goids = set(go for go, d in self.go2dct.items() if 'synonym' in d)
        for goid in goids:
            goobj = self.go2obj[goid]
            ntsyns = getattr(goobj, 'synonym', None)
            assert ntsyns is not None, "{GO} MISSING SYNONYM".format(GO=goid)
            # Iterate through list of synonym data stored in named tuples
            for ntsyn in ntsyns:
                if prt is not None:
                    prt.write("{GO} {NT}\n".format(GO=goid, NT=ntsyn))
                assert ntsyn.text, "SYNONYM CANNOT BE EMPTY"
                assert ntsyn.scope in self.exp_scopes, "INVALID SYNONYM SCOPE"
                for dbxref in ntsyn.dbxrefs:
                    assert self.exp_xrefpat.match(dbxref), "INVALID SYNONYM DBXREF"

    def _get_fldmrk(self, fld):
        """Get a mark for each field indicating if it is required or optional"""
        #pylint: disable=too-many-return-statements
        if fld in self.attrs_req:
            return 'REQ'
        if fld == 'def':
            return 'str'
        if fld in self.attrs_scalar:
            return 'str'
        if fld in self.attrs_set:
            return 'set'
        if fld == 'relationship':
            return 'rel'
        if fld == 'synonym':
            return 'syn'
        if fld == 'xref':
            return 'xrf'
        raise RuntimeError("UNEXPECTED FIELD({})".format(fld))

    def _prt_perc(self, num_rel, name, prt=sys.stdout):
        """Print percentage of GO IDs that have a specific relationship."""
        prt.write("    {N:6,} of {M:,} ({P:3.0f}%) GO IDs has field({A})\n".format(
            N=num_rel, M=self.num_tot, P=float(num_rel)/self.num_tot*100, A=name))

    def _get_cnts_max(self):
        """Get the maximum count of times a specific relationship was seen on a GO."""
        fld2qtys = cx.defaultdict(set)
        flds = self.dcts['flds']
        for recdct in self.go2dct.values():
            for opt in flds:
                if opt in recdct:
                    fld2qtys[opt].add(len(recdct[opt]))
        return {f:max(qtys) for f, qtys in fld2qtys.items()}

    def _get_cnts_gte1(self):
        """Get counts of if a specific relationship was seen on a GO."""
        ctr = cx.Counter()
        flds = self.dcts['flds']
        for recdct in self.go2dct.values():
            for opt in flds:
                if opt in recdct:
                    ctr[opt] += 1
        return ctr

    def chk_cnt_set(self, opt):
        """For each GO ID, check that actual count of a set attr equals expected count."""
        errpat = "SET EXP({EXP}) ACT({ACT}) {GO}\n{DESC}\n:\nEXP:\n{Es}\n\nACT:\n{As}"
        for goid, dct in self.go2dct.items():
            act_set = getattr(self.go2obj[goid], opt, None)
            if opt in dct:
                exp_num = len(dct[opt])
                act_num = len(act_set)
                assert exp_num == act_num, errpat.format(
                    EXP=exp_num, ACT=act_num, GO=goid,
                    DESC=str(self.go2obj[goid]),
                    Es="\n".join(sorted(dct[opt])),
                    As="\n".join(sorted(act_set)))
            else:
                assert act_set is None

    def chk_relationships(self):
        """Check that all GO IDs that should have relationships do have relationships."""
        opt = 'relationship'
        for goid, dct in self.go2dct.items():
            act_rel2goobjs = getattr(self.go2obj[goid], opt, None)
            if opt in dct:
                exp_goids = set(s.split()[1] for s in dct[opt])
                act_goids = set(goobj.id for goobjs in act_rel2goobjs.values() for goobj in goobjs)
                assert exp_goids == act_goids, "EXP({}) ACT({}) {}:\nEXP({})\nACT({})".format(
                    len(exp_goids), len(act_goids), goid, exp_goids, act_goids)
            else:
                assert act_rel2goobjs is None

    def chk_cnt_go(self, opt):
        """Check that all GO IDs that should have relationships do have relationships."""
        if opt == 'def':
            opt = 'defn'
        num_exp = sum(1 for d in self.go2dct.values() if opt in d)
        go_act = set(o.id for o in self.go2obj.values() if hasattr(o, opt))
        assert num_exp == len(go_act), "GO CNTS({}): EXP({}) ACT({})".format(
            opt, num_exp, len(go_act))

    def _init_go2dct(self):
        """Create a dict of GO fields for use as expected results during test."""
        dagdct = {}
        go2dct = {}
        typedefdct = {}
        flds = set()
        with open(self.obo) as ifstrm:
            rec = {}
            for line in ifstrm:
                line = line.rstrip()
                # End of GO record
                if not line:
                    if rec:  # and option is None or option in rec:
                        # 'Definition' is specified in obo as 'def' and in Python by 'defn'
                        if 'def' in rec:
                            rec['defn'] = rec['def']
                        go2dct[rec['GO']] = rec
                    rec = {}
                else:
                    mtch = self.cmpfld.match(line)
                    if mtch:
                        fld = mtch.group(1)
                        val = mtch.group(2)

                        # Beginning of GO record
                        if fld == "id":
                            assert not rec, "NOW({}) WAS({})".format(line, rec)
                            rec = {'GO':val}
                        # Middle of GO record
                        elif rec:
                            flds.add(fld)
                            if fld not in rec:
                                rec[fld] = set()
                            # Strip comment if it exists
                            loc = val.find(' ! ')
                            if loc != -1:
                                val = val[:loc]
                            # Add value
                            rec[fld].add(val)
        return {'dagdct':dagdct, 'go2dct':go2dct, 'typedefdct':typedefdct, 'flds':flds}

    def _init_dnld_dag(self):
        """If dag does not exist, download it."""
        if not os.path.exists(self.obo):
            download_go_basic_obo(self.obo, loading_bar=None)

    def load_dag(self, opt_fields=None):
        """Run numerous tests for various self.reports."""
        tic = timeit.default_timer()
        dag = GODag(self.obo, opt_fields)
        toc = timeit.default_timer()
        msg = "Elapsed HMS for OBO DAG load: {HMS} OPTIONAL_ATTR({O})\n".format(
            HMS=str(datetime.timedelta(seconds=(toc-tic))), O=opt_fields)
        sys.stdout.write(msg)
        return dag


# Copyright (C) 2010-2018, DV Klopfenstein, H Tang, All rights reserved.



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import os.path as op
import sys
sys.path.insert(0, op.join(op.dirname(__file__), ".."))
from goatools.obo_parser import GODag, GraphEngines


if __name__ == '__main__':

    import optparse
    p = optparse.OptionParser("%prog [obo_file]")
    p.add_option("--description", dest="desc",
                 help="write term descriptions to stdout"
                 " from the obo file specified in args", action="store_true")
    p.add_option("--term", dest="term", help="write the parents and children"
                 " of the query term", action="store", type="string",
                 default=None)
    p.add_option("--engine", default="pygraphviz",
                 choices=GraphEngines,
                 help="Graph plot engine, must be one of {0} [default: %default]".\
                        format("|".join(GraphEngines))
                 )
    p.add_option("--gml", action="store_true",
                 help="Write GML output (for Cytoscape) [default: %default]")
    p.add_option("--disable-draw-parents",
                 action="store_false",
                 dest='draw_parents',
                 help="Do not draw parents of the query term")
    p.add_option("--disable-draw-children",
                 action="store_false",
                 dest='draw_children',
                 help="Do not draw children of the query term")

    p.set_defaults(draw_parents=True)
    p.set_defaults(draw_children=True)

    opts, args = p.parse_args()

    if not len(args):
        obo_file = "go-basic.obo"
    else:
        obo_file = args[0]
        assert os.path.exists(obo_file), "file %s not found!" % obo_file

    g = GODag(obo_file)

    if opts.desc:
        g.write_dag()

    # run a test case
    if opts.term is not None:
        rec = g.query_term(opts.term, verbose=True)
        g.draw_lineage([rec], engine=opts.engine,
                       gml=opts.gml,
                       draw_parents=opts.draw_parents,
                       draw_children=opts.draw_children)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os.path as op

from glob import glob
from setuptools import setup
from setup_helper import SetupHelper


name = "goatools"
classifiers = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    ]

# Use the helper
h = SetupHelper(initfile="goatools/__init__.py", readmefile="README.md")

setup_dir = op.abspath(op.dirname(__file__))
requirements = ['wget'] + [x.strip() for x in
                           open(op.join(setup_dir, 'requirements.txt')).readlines()]

setup(
    name=name,
    version=h.version,
    author=h.author,
    author_email=h.email,
    license=h.license,
    long_description=h.long_description,
    packages=[
        name,
        name + ".godag",
        name + ".test_data",
        name + ".anno",
        name + ".anno.extensions"],
    include_package_data=True,
    package_data={"goatools.test_data.nbt_3102": ["*.*"]},
    scripts=glob('scripts/*.py'),
    classifiers=classifiers,
    url='http://github.com/tanghaibao/goatools',
    description="Python scripts to find enrichment of GO terms",
    install_requires=requirements
)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from __future__ import print_function
import os
import os.path as op
import sys
sys.path.insert(0, op.join(op.dirname(__file__), ".."))
from goatools.obo_parser import GODag
from goatools.mapslim import mapslim
from goatools.associations import read_associations


if __name__ == '__main__':

    import optparse
    p = optparse.OptionParser("%prog [options] go_obo_file goslim_obo_file")
    p.add_option("--term", dest="term", help="a term (association id) to map "
                 "to slim IDs. This can not be used together with "
                 "--association_file", action="store", type="string",
                 default=None)
    p.add_option("--association_file", dest="ass_file_name", action="store",
                 help="the file of protein products and their associations "
                 "to be mapped to GO slim terms. This can not be used "
                 "together with --term", type="string", default=None)
    p.add_option("--slim_out", dest="slim_out", action="store", type="string",
                 default="direct", help="One of `direct` or `all`. Defines "
                 "whether the output should contain all slim terms (all "
                 "ancestors) or only direct slim terms (only direct "
                 "ancestors)")
    opts, args = p.parse_args()

    # check for correct number of arguments
    if len(args) != 2:
        p.print_help()
        sys.exit(1)

    obo_file = args[0]
    assert os.path.exists(obo_file), "file %s not found!" % obo_file

    slim_obo_file = args[1]
    assert os.path.exists(slim_obo_file), "file %s not found!" % slim_obo_file

    # check that either --term or --association_file is set
    if (opts.term is None and opts.ass_file_name is None) \
            or ((opts.term is not None) and (opts.ass_file_name is not None)):
        p.print_help()
        sys.exit(1)

    # check that slim_out is either "direct" or "all" and set according flag
    only_direct = None
    if opts.slim_out == "direct":
        only_direct = True
    elif opts.slim_out == "all":
        only_direct = False
    else:
        p.print_help()
        sys.exit(1)

    # load DAGs
    go_dag = GODag(obo_file)
    goslim_dag = GODag(slim_obo_file)

    # in case a single term is given as input:
    if opts.term:
        if opts.term not in go_dag:
            sys.stderr.write(("term %s not found!\n" % opts.term))
            sys.exit(1)
        direct_anc, all_anc = mapslim(opts.term, go_dag, goslim_dag)
        # output either all or only direct slims, depending on user command
        if only_direct:
            slim_terms_str = ";".join(direct_anc)
        else:
            slim_terms_str = ";".join(all_anc)
        print(slim_terms_str)

    # in case a association file is given as input
    if opts.ass_file_name:
        assert os.path.exists(opts.ass_file_name), ("file %s not found!"
                                                    % opts.ass_file_name)
        assocs = read_associations(opts.ass_file_name)
        for protein_product, go_terms in assocs.items():
            all_direct_anc = set()
            all_covered_anc = set()
            all_all_anc = set()
            for go_term in go_terms:
                if go_term not in go_dag:
                    continue
                direct_anc, all_anc = mapslim(go_term, go_dag, goslim_dag)
                all_all_anc |= all_anc
                # collect all covered ancestors, so the direct ancestors
                # can be calculated afterwards
                all_covered_anc |= (all_anc - direct_anc)
            all_direct_anc = all_all_anc - all_covered_anc
            # output either all or only direct, depending on user command
            if only_direct:
                slim_terms_str = ";".join(all_direct_anc)
            else:
                slim_terms_str = ";".join(all_all_anc)
            print((protein_product + "\t" + slim_terms_str))
